@@ -2,7 +2,7 @@ package Test2::IPC::Driver::Files;
 use strict;
 use warnings;
 
-our $VERSION = '1.302039';
+our $VERSION = '1.302041';
 
 
 BEGIN { require Test2::IPC::Driver; our @ISA = qw(Test2::IPC::Driver) }
@@ -18,6 +18,51 @@ use POSIX();
 use Test2::Util qw/try get_tid pkg_to_file IS_WIN32/;
 use Test2::API qw/test2_ipc_set_pending/;
 
+BEGIN {
+    if (IS_WIN32) {
+        my $max_tries = 5;
+
+        *do_rename = sub {
+            my ($from, $to) = @_;
+
+            my $err;
+            for (1 .. $max_tries) {
+                return (1) if rename($from, $to);
+                $err = "$!";
+                last if $_ == $max_tries;
+                sleep 1;
+            }
+
+            return (0, $err);
+        };
+        *do_unlink = sub {
+            my ($file) = @_;
+
+            my $err;
+            for (1 .. $max_tries) {
+                return (1) if unlink($file);
+                $err = "$!";
+                last if $_ == $max_tries;
+                sleep 1;
+            }
+
+            return (0, "$!");
+        };
+    }
+    else {
+        *do_rename = sub {
+            my ($from, $to) = @_;
+            return (1) if rename($from, $to);
+            return (0, "$!");
+        };
+        *do_unlink = sub {
+            my ($file) = @_;
+            return (1) if unlink($file);
+            return (0, "$!");
+        };
+    }
+}
+
 sub use_shm { 1 }
 sub shm_size() { 64 }
 
@@ -107,10 +152,12 @@ sub drop_hub {
         unless get_tid() == $tid;
 
     if ($ENV{T2_KEEP_TEMPDIR}) {
-        rename($hfile, File::Spec->canonpath("$hfile.complete")) or $self->abort_trace("Could not rename file '$hfile' -> '$hfile.complete'");
+        my ($ok, $err) = do_rename($hfile, File::Spec->canonpath("$hfile.complete"));
+        $self->abort_trace("Could not rename file '$hfile' -> '$hfile.complete': $err") unless $ok
     }
     else {
-        unlink($hfile) or $self->abort_trace("Could not remove file for hub '$hid'");
+        my ($ok, $err) = do_unlink($hfile);
+        $self->abort_trace("Could not remove file for hub '$hid': $err") unless $ok
     }
 
     opendir(my $dh, $tdir) or $self->abort_trace("Could not open temp dir!");
@@ -170,7 +217,11 @@ do so if Test::Builder is loaded for legacy reasons.
     # Write and rename the file.
     my ($ok, $err) = try {
         Storable::store($e, $file);
-        rename($file, $ready) or $self->abort("Could not rename file '$file' -> '$ready'");
+        my ($ok, $err) = do_rename("$file", $ready);
+        unless ($ok) {
+            POSIX::sigprocmask(POSIX::SIG_SETMASK(), $old, POSIX::SigSet->new()) if defined $blocked;
+            $self->abort("Could not rename file '$file' -> '$ready': $err");
+        };
         test2_ipc_set_pending(substr($file, -(shm_size)));
     };
 
@@ -239,10 +290,12 @@ sub cull {
 
         my $complete = File::Spec->canonpath("$full.complete");
         if ($ENV{T2_KEEP_TEMPDIR}) {
-            rename($full, $complete) or $self->abort("Could not rename IPC file '$full', '$complete'");
+            my ($ok, $err) = do_rename($full, $complete);
+            $self->abort("Could not rename IPC file '$full', '$complete': $err") unless $ok;
         }
         else {
-            unlink($full) or $self->abort("Could not unlink IPC file: $file");
+            my ($ok, $err) = do_unlink("$full");
+            $self->abort("Could not unlink IPC file '$file': $err") unless $ok;
         }
     }
 
@@ -306,7 +359,8 @@ sub DESTROY {
             $full =~ m/^(.*)$/;
             $full = $1; # Untaint it
             next if $ENV{T2_KEEP_TEMPDIR};
-            unlink($full) or $self->abort("Could not unlink IPC file: $full");
+            my ($ok, $err) = do_unlink($full);
+            $self->abort("Could not unlink IPC file '$full': $err") unless $ok;
             next;
         }
 
